class InMemoryISX:
    def __init__(self, file_system=None):
        self._file_system = file_system

    def preprocess(
            self,
            input_movie_files,
            output_movie_files,
            temporal_downsample_factor=1,
            spatial_downsample_factor=1,
            crop_rect=None,
            crop_rect_format="tlbr",
            fix_defective_pixels=True,
            trim_early_frames=True
    ):
        for output_file in output_movie_files:
            self._file_system.write(output_file, "")

    def spatial_filter(
            self,
            input_movie_files,
            output_movie_files,
            low_cutoff=0.005,
            high_cutoff=0.5,
            retain_mean=False,
            subtract_global_minimum=True
    ):
        for output_file in output_movie_files:
            self._file_system.write(output_file, "")

    def motion_correct(
            self,
            input_movie_files,
            output_movie_files,
            max_translation=20,
            low_bandpass_cutoff=0.004,
            high_bandpass_cutoff=0.016,
            roi=None,
            reference_segment_index=0,
            reference_frame_index=0,
            reference_file_name='',
            global_registration_weight=1,
            output_translation_files=None,
            output_crop_rect_file=None,
            preserve_input_dimensions=False
    ):
        for output_file in output_movie_files:
            self._file_system.write(output_file, "")
        for output_file in output_translation_files or []:
            self._file_system.write(output_file, "")
        self._file_system.write(output_crop_rect_file, "")

    def project_movie(
            self,
            input_movie_files,
            output_image_file,
            stat_type='mean'
    ):
        self._file_system.write(output_image_file, "")

    def dff(
            self,
            input_movie_files,
            output_movie_files,
            f0_type='mean'
    ):
        for output_file in output_movie_files:
            self._file_system.write(output_file, "")

    def pca_ica(
            self,
            input_movie_files,
            output_cell_set_files,
            num_pcs,
            num_ics=120,
            unmix_type='spatial',
            ica_temporal_weight=0,
            max_iterations=100,
            convergence_threshold=0.00001,
            block_size=1000,
            auto_estimate_num_ics=False,
            average_cell_diameter=13
    ):
        for output_file in output_cell_set_files:
            self._file_system.write(output_file, "")

    def event_detection(
            self,
            input_cell_set_files,
            output_event_set_files,
            threshold=5,
            tau=0.2,
            event_time_ref='beginning',
            ignore_negative_transients=True,
            accepted_cells_only=False
    ):
        for output_file in output_event_set_files:
            self._file_system.write(output_file, "")

    def auto_accept_reject(
            self,
            input_cell_set_files,
            input_event_set_files,
            filters=None
    ):
        pass

    def longitudinal_registration(
        self,
        input_cell_set_files,
        output_cell_set_files,
        input_movie_files=[],
        output_movie_files=[],
        csv_file='',
        min_correlation=0.5,
        accepted_cells_only=False,
        transform_csv_file='',
        crop_csv_file=''
    ):
        for output_file in output_cell_set_files:
            self._file_system.write(output_file, "")
        for output_file in output_movie_files:
            self._file_system.write(output_file, "")
        if csv_file:
            self._file_system.write(csv_file, "")
        if transform_csv_file:
            self._file_system.write(transform_csv_file, "")
        if crop_csv_file:
            self._file_system.write(crop_csv_file, "")

    @property
    def CellSet(self):
        class CellSet:
            @staticmethod
            def read(p):
                if not self._file_system.exists(p):
                    raise IOError(f"Cannot read file: {p}")
                class Dummy:
                    @property
                    def num_cells(self):
                        return 1
                return Dummy()
        return CellSet

    def export_movie_to_tiff(self,
                             input_movie_files,
                             output_movie_file,
                             write_invalid_frames=False,
                             ):

        for input_file in input_movie_files:
            self._file_system.write(output_movie_file, "")

    def export_movie_to_nwb(self,
                            input_movie_files,
                            output_movie_file,
                            write_invalid_frames=False,
                            ):
        for input_file in input_movie_files:
            self._file_system.write(output_movie_file, "")

    def make_output_file_path(
            self,
            in_file,
            out_dir,
            suffix,
            ext="isxd"
    ):
        base = self._file_system.base_path(in_file)
        stem, _ = self._file_system.split_text(base)
        if suffix:
            stem = f"{stem}-{suffix}"
        new_filename = f"{stem}.{ext}"
        return self._file_system.join(out_dir, new_filename)

    def make_output_file_paths(
            self,
            in_files,
            out_dir,
            suffix,
            ext="isxd"
    ):

        return [
            self.make_output_file_path(in_file, out_dir, suffix, ext)
            for in_file in in_files
        ]



from ci_pipe.decorators import step
from ci_pipe.errors.caiman_backend_not_configured_error import CaimanBackendNotConfiguredError


class CaimanModule:
    MOTION_CORRECTION_STEP = "Caiman Motion Correction"
    MOTION_CORRECTION_VIDEOS_SUFFIX = "MC"

    def __init__(self, caiman, ci_pipe):
        if caiman is None:
            raise CaimanBackendNotConfiguredError()
        self._caiman = caiman
        self._ci_pipe = ci_pipe

    @step(MOTION_CORRECTION_STEP)
    def motion_correction(
            self,
            inputs,
            *,
            caiman_strides=(48, 48),
            caiman_overlaps=(24, 24),
            caiman_max_shifts=(6, 6),
            caiman_max_deviation_rigid=3,
            caiman_pw_rigid=True,
            caiman_shifts_opencv=True,
            caiman_border_nan='copy',
            caiman_save_movie=True
    ):
        # TODO: Think if we should grab all potential extensions accepted by motion correction
        output = []
        output_dir = self._ci_pipe.create_output_directory_for_next_step(self.MOTION_CORRECTION_STEP)

        for input_data in inputs('videos-tif'):
            motion_correct_handler = self._caiman.motion_correction.MotionCorrect(
                fname=input_data['value'],
                strides=caiman_strides,
                overlaps=caiman_overlaps,
                max_shifts=caiman_max_shifts,
                max_deviation_rigid=caiman_max_deviation_rigid,
                pw_rigid=caiman_pw_rigid,
                shifts_opencv=caiman_shifts_opencv,
                border_nan=caiman_border_nan,
            )
            motion_correct_handler.motion_correct(save_movie=caiman_save_movie)
            mmap_files = motion_correct_handler.mmap_file
            mmap_path = mmap_files[0] # we are processing only one at a time, that's why we can unpack it like this

            memmapped_movie = self._caiman.load(mmap_path)

            tif_output_path = self._ci_pipe.make_output_file_path(
                mmap_path,
                output_dir,
                self.MOTION_CORRECTION_VIDEOS_SUFFIX,
                ext="tif",
            )

            memmapped_movie.save(tif_output_path)
            output.append({'ids': input_data['ids'], 'value': tif_output_path})

        return {"videos-tif": output}



class Step:
    def __init__(self, step_name, look_up_function=None, step_function=None, args=None, kwargs=None, step_outputs=None):
        self._step_name = step_name
        self._step_function = step_function
        self._args = args if args is not None else []
        self._kwargs = kwargs if kwargs is not None else {}

        # TODO: Handle step_outputs more gracefully?
        if step_outputs is not None:
            self._step_outputs = step_outputs
        elif self._step_function is not None:
            self._step_outputs = self._step_function(look_up_function, *self._args, **self._kwargs)
        else:
            self._step_outputs = {}

    @classmethod
    def from_dict(cls, data):
        if data is None:
            return None
        return cls(
            data.get("name"),
            look_up_function=None,
            step_function=None,  # Don't execute the function during deserialization
            args=None,
            kwargs=data.get("params"),
            step_outputs=data.get("outputs"),
        )

    @classmethod
    def restored_from_trace(cls, name, outputs, params):
        obj = cls.__new__(cls)
        obj._name = name
        obj._output_lookup = None
        obj._fn = None
        obj._args = ()
        obj._kwargs = params or {}
        obj._step_outputs = outputs  # Preload, do not execute
        return obj

    def step_output(self):
        return self._step_outputs

    def name(self):
        return self._step_name

    def output(self):
        return self._step_outputs

    def arguments(self):
        return self._kwargs

    def to_dict(self):
        return {
            "name": self.name(),
            "outputs": self.output(),
            "params": self.arguments()
        }



import functools
import inspect
from inspect import Parameter, Signature
from typing import TypeVar, Callable, Protocol, Any, TYPE_CHECKING

try:
    from typing import ParamSpec
except ImportError:
    from typing_extensions import ParamSpec

if TYPE_CHECKING:
    from .pipeline import CIPipe

T_Pipeline = TypeVar("T_Pipeline", bound="CIPipe")
P = ParamSpec("P")

class PipelineBound(Protocol[T_Pipeline]):
    _ci_pipe: T_Pipeline


def step(step_name: str) -> Callable[[Callable[..., dict[str, list]]], Callable[..., "CIPipe"]]:
	"""
	Method decorator for pipeline steps without requiring a nested function.

	Usage:
		@step("My Step Name")
		def my_method(self, inputs, *, opt=1):

	"""

	def decorator(method: Callable[..., dict[str, list]]) -> Callable[..., "CIPipe"]:
		@functools.wraps(method)
		def wrapper(self: PipelineBound["CIPipe"], *call_args: Any, **call_kwargs: Any) -> "CIPipe":
			if not hasattr(self, "_ci_pipe") or self._ci_pipe is None:
				raise RuntimeError("Decorator @step requires 'self._ci_pipe' to be set.")

			bound_method = method.__get__(self, type(self))

			def step_function(inputs, *s_args, **s_kwargs):
				return bound_method(inputs, *s_args, **s_kwargs)

			orig_sig = inspect.signature(method)
			new_params = []
			for name, p in orig_sig.parameters.items():
				if name == 'self':
					continue
				if name == 'inputs':
					new_params.append(Parameter('inputs', kind=Parameter.POSITIONAL_OR_KEYWORD))
				else:
					default = p.default if p.default is not inspect._empty else inspect._empty
					new_params.append(Parameter(name, kind=Parameter.KEYWORD_ONLY, default=default))
			step_function.__signature__ = Signature(new_params)

			return self._ci_pipe.step(step_name, step_function, *call_args, **call_kwargs)

		wrapper.__signature__ = inspect.signature(method)
		wrapper.__doc__ = method.__doc__
		wrapper.__annotations__ = method.__annotations__.copy()
		wrapper.__annotations__["return"] = "CIPipe"

		return wrapper

	return decorator



class Pipeline:
    def __init__(self, inputs, defaults, outputs_directory):
        self.inputs = inputs
        self.defaults = defaults
        self.outputs_directory = outputs_directory

    @classmethod
    def from_dict(cls, data):
        return cls(
            data.get("inputs", {}),
            data.get("defaults", {}),
            data.get("outputs_directory")
        )

    def to_dict(self):
        return {
            "inputs": self.inputs,
            "defaults": self.defaults,
            "outputs_directory": self.outputs_directory
        }



import json

from ci_pipe.trace.ci_pipe_trace import CIPipeTrace


class TraceRepository:
    def __init__(self, file_system, filename, validator=None):
        self._file_system = file_system
        self._filename = filename
        self._validator = validator

    def load(self) -> CIPipeTrace:
        try:
            json_trace = json.loads(self._file_system.read(self._filename))
        except Exception:
            json_trace = {}
        return CIPipeTrace.from_dict(json_trace)

    def save(self, trace: CIPipeTrace):
        trace_as_json = trace.to_dict()
        self._file_system.write(self._filename, json.dumps(trace_as_json, indent=4))

    def exists(self):
        return self._file_system.exists(self._filename)

    # TODO: Think if we need to handle this onSave instead of here
    def validate(self):
        if not self._validator:
            return True
        data_as_json = self.load().to_dict()
        return self._validator.validate(data_as_json)